@@ -1,7 +1,7 @@
 #! /usr/bin/perl -w
 
-# Copyright (c) 2003-2011 Henk Penning, all rights reserved.
-# penning@cs.uu.nl, http://www.cs.uu.nl/staff/henkp.html
+# Copyright (c) 2003-2012 Henk Penning, all rights reserved.
+# penning@uu.nl, http://www.cs.uu.nl/staff/henkp.html
 # Version 1.1 was donated to the Apache Software Foundation 2003 Jan 28.
 # Permission is hereby granted, free of charge, to any person obtaining a
 # copy of this software and associated documentation files (the "Software"),
@@ -28,16 +28,16 @@
 use strict ;
 
 our $PRG = 'mirmon' ;
-our $VER = "2.4" ;
+our $VER = "2.7" ;
 
 our $DEF_TIMEOUT = 300 ;
 our $HIST        = 14 ;
 our $TIM_PAT     = '^(\d+)([smhd])$' ;
-our %APA_TYPES   = () ; for ( qw(backup ftp http) ) { $APA_TYPES { $_ } ++ ; }
-our %GET_OPTS    = () ; for ( qw(all update) )      { $GET_OPTS  { $_ } ++ ; }
+our %APA_TYPES   = () ; $APA_TYPES { $_ } ++ for qw(backup ftp http rsync) ;
+our %GET_OPTS    = () ; $GET_OPTS  { $_ } ++ for qw(all update) ;
 our $HIST_DELTA  = 24 * 60 * 60 ;
-our $APRX_DELTA  = 60 ;
-our $HOME        = 'http://www.cs.uu.nl/people/henkp/mirmon/' ;
+our $APRX_DELTA  = 300 ;
+our $HOME        = 'http://www.staff.science.uu.nl/~penni101/mirmon/' ;
 
 package Base ; #####################################################
 
@@ -225,6 +225,7 @@ sub get_config
 sub get_state
   { my $self = shift ;
     my $conf = $self -> conf ;
+    my $name = $conf -> project_name ;
     my $state = $conf -> state ;
     my $res = {} ;
     open STATE, $state or die "can't open $state ($!)" ;
@@ -238,6 +239,7 @@ sub get_state
     my $mlist = $conf -> mirror_list ;
     my $style = $conf -> list_style ;
     my %in_list = () ;
+    my $changes = '' ;
     open MLIST, $mlist or die "can't open $mlist ($!)" ;
     for my $line ( <MLIST> )
       { chop $line ;
@@ -264,7 +266,7 @@ sub get_state
         $in_list { $url } ++ ;
 
         unless ( exists $res -> { $url } )
-          { printf "*** added to list %s\n", $url unless Mirmon::quiet ;
+          { $changes .= sprintf "added %s\n", $url unless Mirmon::quiet ;
             $res -> { $url } = Mirmon::Mirror -> init ( $self, $url ) ;
           }
         my $mirror = $res -> { $url } ;
@@ -276,10 +278,12 @@ sub get_state
     for my $url ( sort keys %$res )
       { # printf "%s\n", $res -> { $url } -> state ;
         unless ( exists $in_list { $url } )
-          { printf "*** removed from list %s\n", $url unless Mirmon::quiet ;
+          { $changes .= sprintf "removed %s\n", $url unless Mirmon::quiet ;
             delete $res -> { $url } ;
           }
       }
+    printf "changes in mirror-list for '%s':\n%s", $name, $changes
+      if $changes ;
     $self -> state ( $res ) ;
   }
 
@@ -309,73 +313,172 @@ sub get_regions
       { chop ;
         next if /^#/ ;
         my ( $code, $dash, $reg ) = split ' ', $_, 3 ;
-        $self -> { regions } { lc $code } = lc $reg ;
+        $self -> { regions } { lc $code } = $reg ;
       }
     close REGS ;
   }
 
+sub _cmp_ccs
+  { my $ccs = shift ;
+    my $x  = shift ;
+    my $y  = shift ;
+    my $xx = $ccs -> { $x } ;
+    my $yy = $ccs -> { $y } ;
+    if ( ! defined $xx and ! defined $yy )
+      { $x cmp $y ; }
+    elsif ( ! defined $xx )
+      { -1 ; }
+    elsif ( ! defined $yy )
+      { +1 ; }
+    else
+      { $xx cmp $yy ; }
+  }
+
+sub _pr_round
+  { my $x = shift ;
+    my $i = int $x ;
+    my $f = $x - $i ;
+    $i + ( rand 1 < $f ? 1 : 0 ) ;
+  }
+
+sub _diag_qs
+  { my $qs = shift ;
+    join ', ', map { sprintf "%s %s" , $_, scalar @{ $qs -> { $_ } } ; }
+      sort keys %$qs ;
+  }
+
+sub _rpick
+  { my $row = shift ;
+    die "_rpick : row empty" unless @$row ;
+    my $idx = int rand @$row ;
+    my $res = $row -> [ $idx ] ;
+    $row -> [ $idx ] = $row -> [ $#{$row} ] ;
+    pop @$row ;
+    $res ;
+  }
+
+sub _buck_split
+  { my $que = shift ;
+    my $tmp = [] ;
+    for my $mirr ( @$que )
+      { my $lp = $mirr -> last_probe ;
+        my $hr = int ( ( $^T - $lp ) / 60 / 60 + 0.5 ) ;
+        push @{ $tmp -> [ $hr ] }, $mirr ;
+      }
+    [ grep defined $_, @$tmp ] ;
+  }
+
+sub _buck_join
+  { my $bucks = shift ;
+    my $res = [] ;
+    push @$res, @$_ for @$bucks ;
+    $res ;
+  }
+
+sub _buck_pick
+  { my $bucks = shift ;
+    die "buck_pick : bucks empty" unless @$bucks ;
+    my $buck = ( sort { @$b <=> @$a } @$bucks ) [ 0 ] ;
+    _rpick $buck ;
+  }
+
+sub _randomize
+  { my $ques = shift ;
+    my $poll = shift ;
+    my $hrs  = int ( $poll / 60 / 60 + 0.5 ) ;
+
+    my $diag1 = _diag_qs $ques ;
+
+    my $todos = $ques -> { todo } ;
+    my $dones = $ques -> { done } ;
+    my $cnt   = @$todos + @$dones ;
+    my $avg   = $hrs ? $cnt / $hrs : 0 ;
+    my $iavg  = _pr_round $avg ;
+    my $pick  = 0 ;
+    my $bucks = _buck_split $dones ;
+
+    while ( @$todos < $iavg and $pick < @$dones )
+      { push @$todos, _buck_pick $bucks ;
+        $pick ++ ;
+      }
+
+    $ques -> { done } = _buck_join $bucks ;
+
+    sprintf ''
+      . "  hrs %s, %s\n"
+      . "  avg %.2f -> %d , picked %d ; queued %s\n"
+      . "  hrs %s, %s\n"
+      , $hrs, $diag1
+      , $avg, $iavg, $pick, scalar @$todos
+      , $hrs, _diag_qs ( $ques )
+      ;
+  }
+
 sub get_dates
   { my $self  = shift ;
     my $get   = shift ;
     my $state = $self -> state ;
-    my $conf = $self -> conf ;
-    my $CMD  = $conf -> probe ;
-    my $PAR  = $conf -> max_probes ;
-    my %m4h  = () ;
-    my @QUE  = () ;
-    my @NOQ  = () ;
-    my $GET = IO::Select -> new () ;
+    my $conf  = $self -> conf ;
+    my $CMD   = $conf -> probe ;
+    my $PAR   = $conf -> max_probes ;
+    my %m4h   = () ;
+    my @QUE   = () ;
+    my $GET   = IO::Select -> new () ;
+    my $ques  = {} ;
+    for my $col ( qw(new red grn xtr) )
+      { $ques -> { $col } { $_ } = [] for qw(done todo) ; }
+    my $max_poll = s4tim $conf -> max_poll ;
+    my $min_poll = s4tim $conf -> min_poll ;
 
-    my $cnt = 0 ;
-    my $nok = 0 ;
+    if ( Mirmon::verbose ) { printf "mirrors %d\n", scalar keys %$state ; }
 
-    for my $url ( sort keys %$state )
-      { my $mirror = $state -> { $url } ;
-        $cnt ++ if $mirror -> last_status eq 'ok' ;
-        if ( $get eq 'all' or $mirror -> last_probe eq 'undef' )
-          { push @QUE, $mirror ; }
-        elsif ( $get eq 'update' )
-          { my $stat = $mirror -> last_status ;
+    if ( $get eq 'all' )
+      { @QUE = sort { $a -> url cmp $b -> url } values %$state ; }
+    elsif ( $get eq 'update' )
+      { my $maxp = $^T - $max_poll ;
+        my $minp = $^T - $min_poll ;
+
+if ( Mirmon::verbose )
+  { printf "max_poll %s\n", scalar localtime $maxp ;
+    printf "min_poll %s\n", scalar localtime $minp ;
+  }
+        for my $url ( sort keys %$state )
+          { my $mirror = $state -> { $url } ;
+            my $stat = $mirror -> last_status ;
             my $vrfy = $mirror -> last_ok_probe ;
             my $lprb = $mirror -> last_probe ;
-            if ( aprx_le $lprb, $^T - s4tim $conf -> min_poll )
-              { if ( $stat ne 'ok' )
-                  { push @QUE, $mirror ; $nok ++ ; }
-                elsif ( aprx_le $vrfy, $^T - s4tim $conf -> max_poll )
-                  { push @QUE, $mirror ; }
-                else
-                  { push @NOQ, $mirror ; }
+            my $col ;
+            my $que ;
+            if ( $stat eq 'undef' ) # never probed ; new mirror ; todo
+              { $col = 'new' ; $que = 'todo' ; }
+            elsif ( $conf -> get_xtr ( $mirror -> region ) )
+              { $col = 'xtr' ; $que = 'todo' ; }
+            else
+              { my $poll = $stat eq 'ok' ? $maxp : $minp ;
+                $col     = $stat eq 'ok' ? 'grn' : 'red' ;
+                $que     = ( aprx_le $lprb, $poll ) ? 'todo' : 'done' ;
               }
-          }
-        else
-          { die "unknown opt_get '$get'" ; }
-      }
-
-    if ( Mirmon::verbose )
-      { my $que = scalar @QUE ; my $noq = scalar @NOQ ;
-        printf "ok mirrors %d, queued %d, not queued %d, ok %d, nok %d\n"
-          , $cnt, $que, $noq, $que - $nok, $nok
-      }
-
-    if ( $conf -> randomize )
-      { my $hrs  = int ( ( s4tim $conf -> max_poll ) / 60 / 60 + 0.5 ) ;
-        my $avg  = int ( $cnt / $hrs + 0.5 ) ;
-        my $prc  = ( scalar keys %$state ) / 50 ;
-        my $flr  = int $prc ;
-        my $extras = $flr + ( rand 1 < ( $prc - $flr ) ) ;
-        my $picked = 0 ;
-
-        while ( @QUE < $avg + $nok and @NOQ and $picked < $extras )
-          { my $idx = int rand @NOQ ;
-            push @QUE, $NOQ [ $idx ] ;
-            $NOQ [ $idx ] = $NOQ [ $#NOQ ] ;
-            pop @NOQ ;
-            $picked ++ ;
+            push @{ $ques -> { $col } { $que } }, $mirror ;
           }
 
-        printf "avg mirrors/hr %d, max extras %d, picked %d ; queued %s\n"
-          , $avg, $extras, $picked, scalar @QUE if Mirmon::verbose ;
+        if ( $conf -> randomize )
+          { my $msg = "randomize green\n" ;
+            $msg .= _randomize $ques -> { grn }, $max_poll ;
+            $msg .= "randomize red\n" ;
+            $msg .= _randomize $ques -> { red }, $min_poll ;
+            print $msg if Mirmon::verbose ;
+          }
+        @QUE =
+          ( @{ $ques -> { new } { todo } }
+          , @{ $ques -> { red } { todo } }
+          , @{ $ques -> { grn } { todo } }
+          , @{ $ques -> { xtr } { todo } }
+          ) ;
       }
+    else
+      { die "unknown opt_get '$get'" ; }
+    
+    if ( Mirmon::verbose ) { printf "queued %d\n\n", scalar @QUE ; }
 
     while ( @QUE )
       { my $started = 0 ;
@@ -871,7 +974,8 @@ sub gen_histogram
       { $res .= sprintf "<BR>each %s unit represents %s mirror sites.\n"
           , $units, sprintf ( "%.1f", $max / $H ) ;
       }
-    return H2 ( 'age histogram' ) . BQ $res ;
+    return H2 ( NAM 'age-histogram', 'age histogram' ) 
+      . BQ $res ;
   }
 
 sub gen_page
@@ -1020,10 +1124,11 @@ $histo_top
 HEAD
 
     for my $reg
-      ( sort
-       { ( $CCS -> { $a } || $a ) cmp ( $CCS -> { $b } || $b ) ; }
-       keys %tab
-      )
+      ( sort { _cmp_ccs $CCS, $a, $b } keys %tab )
+#      { ( $CCS -> { $a } ? lc ( $CCS -> { $a } ) : $a )
+#    cmp ( $CCS -> { $b } ? lc ( $CCS -> { $b } ) : $b )
+#      } keys %tab
+#     )
       { my $mirrors = $tab { $reg } ;
 
         my $ccs = exists $CCS -> { $reg } ? $CCS -> { $reg } : $reg ;
@@ -1038,8 +1143,8 @@ HEAD
         for my $mirror ( sort { $a -> cmp ( $b ) } @$mirrors )
           { print "<TR>\n" ;
             printf "  <TD ALIGN=RIGHT>%s&nbsp;&nbsp;%s</TD>\n  <TD>%s</TD>\n"
-              , $mirror -> site_url
-              , $mirror -> home_url
+              , $mirror -> url_site
+              , $mirror -> url_home
               , $mirror -> type
               ;
 
@@ -1118,6 +1223,7 @@ our %CNF_defaults =
   , htm_top      => ''
   , htm_foot     => ''
   , htm_head     => ''
+  , always_get   => ''
   ) ;
 
 our @REQ_KEYS =
@@ -1132,6 +1238,12 @@ my @PUT_HGRAM  = qw(top bottom nowhere) ;
 
 eval Base -> mk_methods ( keys %CNF_KEYS, qw(root site_url) ) ;
 
+sub get_xtr
+  { my $self = shift ;
+    my $reg  = shift ;
+    scalar grep { $_ eq $reg } split ' ', $self -> always_get ;
+  }
+
 sub new
   { my $self = shift ;
     my $FILE = shift ;
@@ -1293,15 +1405,18 @@ sub state_history_hist
 sub _parse
   { my $self = shift ;
     my $url  = $self -> url ;
-    my ( $type, $site, $home ) ;
+    my ( $type, $site, $home, $path ) ;
     if ( $url =~ m!^(ftp|https?|rsync)://([^/:]+)(:\d+)?/! )
-      { $type = $1 ; $site = $2 ; $home = $& ; }
-    return $type, $site, $home ;
+      { $type = $1 ; $site = $2 ; $home = $& ; $path = $' ; }
+    else
+      { warn "can't parse url ($url)" ; }
+    return $type, $site, $home, $path ;
   }
 
 sub type { my $self = shift ; ( $self -> _parse ) [ 0 ] ; }
 sub site { my $self = shift ; ( $self -> _parse ) [ 1 ] ; }
 sub home { my $self = shift ; ( $self -> _parse ) [ 2 ] ; }
+sub path { my $self = shift ; ( $self -> _parse ) [ 3 ] ; }
 
 sub age_in_days
   { my $self = shift ;
@@ -1463,8 +1578,26 @@ sub _url
     $hrf =~ /^rsync/ ? $txt : URL $hrf, $txt ;
   }
 
-sub site_url { my $self = shift ; _url $self -> url , $self -> site ; }
-sub home_url { my $self = shift ; _url $self -> home, '@' ; }
+sub url_site
+  { my $self = shift ;
+    my $type = $self -> type ;
+    if ( $type eq 'rsync' )
+      { my $path = $self -> path ;
+        chop $path if $path =~ m!/$! ;
+        sprintf '%s::%s', $self -> site , $path ;
+      }
+    else
+      { URL $self -> url , $self -> site ; }
+  }
+
+sub url_home
+  { my $self = shift ;
+    my $type = $self -> type ;
+    if ( $type eq 'rsync' )
+      { '@' ; }
+    else
+      { URL $self -> home, '@' ; }
+  }
 
 =pod
 
@@ -1698,30 +1831,30 @@ mirmon(1)
 
 =begin html
 
-  &copy; 2003-2011
-  <A HREF="http://people.cs.uu.nl/henkp/">Henk P. Penning</A>,
-  <A HREF="http://www.cs.uu.nl/">Computer Science Department</A>,
+  &copy; 2003-2012
+  <A HREF="http://www.staff.science.uu.nl/~penni101/">Henk P. Penning</A>,
+  <A HREF="http://www.uu.nl/faculty/science/EN/">Faculty of Science</A>,
   <A HREF="http://www.uu.nl/">Utrecht University</A>
   <BR>
-  mirmon-2.4 - Mon Jan 17 18:14:46 2011 ; henkp
+  mirmon-2.7 - Sat Dec 15 08:43:36 2012 ; henkp
 
 =end html
 
 =begin man
 
-  (c) 2003-2011 Henk P. Penning
-  Computer Science Department, Utrecht University
-  http://people.cs.uu.nl/henkp/ -- penning@cs.uu.nl
-  mirmon-2.4 - Mon Jan 17 18:14:46 2011 ; henkp
+  (c) 2003-2012 Henk P. Penning
+  Faculty of Science, Utrecht University
+  http://www.staff.science.uu.nl/~penni101/ -- penning@uu.nl
+  mirmon-2.7 - Sat Dec 15 08:43:36 2012 ; henkp
 
 =end man
 
 =begin text
 
-  (c) 2003-2011 Henk P. Penning
-  Computer Science Department, Utrecht University
-  http://people.cs.uu.nl/henkp/ -- penning@cs.uu.nl
-  mirmon-2.4 - Mon Jan 17 18:14:46 2011 ; henkp
+  (c) 2003-2012 Henk P. Penning
+  Faculty of Science, Utrecht University
+  http://www.staff.science.uu.nl/~penni101/ -- penning@uu.nl
+  mirmon-2.7 - Sat Dec 15 08:43:36 2012 ; henkp
 
 =end text
 
